import typing
import argparse
import threading
import json
import socket
import contextlib
import time
import agutil.io

class SetupManager(object):
    """
    Manages setup/teardown for Canine jobs.
    Runs a small socketserver to provide setup and teardown data.
    """
    def __init__(self):
        """
        Initializes the Manager
        """
        self.setup = {}
        self.teardown = {}

    def add_job(self, jobId: str, setup: str, teardown: str):
        """
        Adds the given setup and teardown data for the job
        """
        self.setup[jobId] = setup
        self.teardown[jobId] = teardown

    def export(self, filepath_or_handle: typing.Union[str, typing.TextIO]):
        """
        Dumps the SetupManager to a json file, in a format meant for import on
        remote system
        """
        with contextlib.ExitStack() as stack:
            if isinstance(filepath_or_handle, str):
                filepath_or_handle = stack.enter_context(open(filepath_or_handle, 'w'))
            json.dump({
                'setup': self.setup,
                'teardown': self.teardown
            }, filepath_or_handle)

    def _handle_connection(self, socket, lock):
        try:
            cmd = socket.recv(decode=True)
            if cmd == 'setup':
                jid = socket.recv(decode=True)
                if jid in self.setup:
                    socket.send(self.setup[jid])
                    with lock:
                        self.__incomplete.add(jid)
                else:
                    socket.send('::ERR:Job ID {} not found'.format(jid))
            elif cmd == 'teardown':
                jid = socket.recv(decode=True)
                if jid in self.teardown:
                    socket.send(self.teardown[jid])
                    with lock:
                        self.__incomplete.remove(jid)
                else:
                    socket.send('::ERR:Job ID {} not found'.format(jid))
            else:
                socket.send('::ERR:Command {} not supported'.format(cmd))
        finally:
            socket.close()

    def serve(self, port: int):
        """
        Begins serving the setup/teardown data.
        RUNS IN FOREGROUND! Run in a separate thread if you do not want to block python.
        Automatically shuts down after all jobs have completed teardown
        """
        lock = threading.RLock()
        self.__incomplete = {*self.teardown}
        server = agutil.io.SocketServer(port)
        try:
            server.sock.settimeout(300)
            last_connection = time.monotonic()
            while len(self.__incomplete):
                try:
                    threading.Thread(
                        target=self._handle_connection,
                        args=(server.accept(), lock),
                        daemon=True
                    ).start()
                    last_connection = time.monotonic()
                except socket.timeout:
                    if (time.monotonic() - last_connection) > 3600:
                        # If the server has been idle for 1 hour, shut down
                        return
        finally:
            server.close()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        'canine-manager',
        description='Setup and teardown manager for canine jobs'
    )

    parser.add_argument(
        'action',
        choices=['serve', 'setup', 'teardown']
    )

    parser.add_argument(
        'port',
        type=int
    )

    parser.add_argument(
        'target'
    )

    args = parser.parse_args()

    if args.action == 'serve':
        mgr = SetupManager()
        with open(args.target) as r:
            config = json.load(r)
        keys = {key for key in config['setup']} | {key for key in config['teardown']}
        for key in keys:
            mgr.add_job(
                key,
                config['setup'][key] if key in config['setup'] else '',
                config['teardown'][key] if key in config['teardown'] else ''
            )
        mgr.serve(args.port)
    else:
        socket = agutil.io.Socket('localhost', args.port)
        socket.send(args.action)
        socket.send(args.target)
        msg = socket.recv(decode=True)
        if msg.startswith('::ERR:'):
            raise ValueError(msg[6:])
        print(msg)



import typing
import os
import time
import sys
import warnings
import traceback
from subprocess import CalledProcessError
from .adapters import AbstractAdapter, ManualAdapter, FirecloudAdapter
from .backends import AbstractSlurmBackend, LocalSlurmBackend, RemoteSlurmBackend, TransientGCPSlurmBackend, TransientImageSlurmBackend
from .localization import AbstractLocalizer, BatchedLocalizer, LocalLocalizer, RemoteLocalizer, NFSLocalizer
from .utils import check_call
import yaml
import pandas as pd
from agutil import status_bar
version = '0.6.0'

ADAPTERS = {
    'Manual': ManualAdapter,
    'Firecloud': FirecloudAdapter,
    'Terra': FirecloudAdapter
}

BACKENDS = {
    'Local': LocalSlurmBackend,
    'Remote': RemoteSlurmBackend,
    'TransientGCP': TransientGCPSlurmBackend,
    'TransientImage': TransientImageSlurmBackend
}

LOCALIZERS = {
    'Batched': BatchedLocalizer,
    'Local': LocalLocalizer,
    'Remote': RemoteLocalizer,
    'NFS': NFSLocalizer
}

ENTRYPOINT = """#!/bin/bash
export CANINE="{version}"
export CANINE_BACKEND="{{backend}}"
export CANINE_ADAPTER="{{adapter}}"
export CANINE_ROOT="{{CANINE_ROOT}}"
export CANINE_COMMON="{{CANINE_COMMON}}"
export CANINE_OUTPUT="{{CANINE_OUTPUT}}"
export CANINE_JOBS="{{CANINE_JOBS}}"
source $CANINE_JOBS/$SLURM_ARRAY_TASK_ID/setup.sh
{{pipeline_script}}
CANINE_JOB_RC=$?
source $CANINE_JOBS/$SLURM_ARRAY_TASK_ID/teardown.sh
exit $CANINE_JOB_RC
""".format(version=version)

class Orchestrator(object):
    """
    Main class
    Parses a configuration object, initializes, runs, and cleans up a Canine Pipeline
    """

    @staticmethod
    def stringify(obj: typing.Any) -> typing.Any:
        """
        Recurses through the dictionary, converting objects to strings
        """
        if isinstance(obj, list):
            return [
                Orchestrator.stringify(elem)
                for elem in obj
            ]
        elif isinstance(obj, dict):
            return {
                key:Orchestrator.stringify(val)
                for key, val in obj.items()
            }
        return str(obj)

    @staticmethod
    def fill_config(cfg: typing.Union[str, typing.Dict[str, typing.Any]]) -> typing.Dict[str, typing.Any]:
        """
        Loads the given config object (or reads from the given filepath)
        Applies Canine defaults, then returns the final config dictionary
        """
        if isinstance(cfg, str):
            with open(cfg) as r:
                cfg = yaml.load(r, Loader=yaml.loader.SafeLoader)
        DEFAULTS = {
            'name': 'canine',
            'adapter': {
                'type': 'Manual',
            },
            'backend': {
                'type': 'Local'
            },
            'localization': {
                'strategy': 'Batched'
            }
        }
        for key, value in DEFAULTS.items():
            if key not in cfg:
                cfg[key] = value
            elif isinstance(value, dict):
                cfg[key] = {**value, **cfg[key]}
        return cfg


    def __init__(self, config: typing.Union[str, typing.Dict[str, typing.Any]]):
        """
        Initializes the Orchestrator from a given config
        """
        config = Orchestrator.fill_config(config)
        self.name = config['name']
        if 'script' not in config:
            raise KeyError("Config missing required key 'script'")
        self.script = config['script']
        if isinstance(self.script, str):
            if not os.path.isfile(self.script):
                raise FileNotFoundError(self.script)
        elif not isinstance(self.script, list):
            raise TypeError("script must be a path to a bash script or a list of bash commands")
        self.raw_inputs = Orchestrator.stringify(config['inputs']) if 'inputs' in config else {}
        self.resources = Orchestrator.stringify(config['resources']) if 'resources' in config else {}
        adapter = config['adapter']
        if adapter['type'] not in ADAPTERS:
            raise ValueError("Unknown adapter type '{type}'".format(**adapter))
        self._adapter_type=adapter['type']
        self.adapter = ADAPTERS[adapter['type']](**{arg:val for arg,val in adapter.items() if arg != 'type'})
        backend = config['backend']
        if backend['type'] not in BACKENDS:
            raise ValueError("Unknown backend type '{type}'".format(**backend))
        self._backend_type = backend['type']
        del backend['type']
        self._slurmconf_path = backend['slurm_conf_path'] if 'slurm_conf_path' in backend else None
        self.backend = BACKENDS[self._backend_type](**backend)
        self.localizer_args = config['localization'] if 'localization' in config else {}
        if self.localizer_args['strategy'] not in LOCALIZERS:
            raise ValueError("Unknown localization strategy '{}'".format(self.localizer_args))
        self._localizer_type = LOCALIZERS[self.localizer_args['strategy']]
        self.localizer_args = {key:val for key,val in self.localizer_args.items() if key != 'strategy'}
        self.localizer_overrides = {}
        if 'overrides' in self.localizer_args:
            self.localizer_overrides = {**self.localizer_args['overrides']}
            del self.localizer_args['overrides']
        self.raw_outputs = Orchestrator.stringify(config['outputs']) if 'outputs' in config else {}
        if len(self.raw_outputs) == 0:
            warnings.warn("No outputs declared", stacklevel=2)
        if 'stdout' not in self.raw_outputs:
            self.raw_outputs['stdout'] = '../stdout'
        if 'stderr' not in self.raw_outputs:
            self.raw_outputs['stderr'] = '../stderr'

    def run_pipeline(self, output_dir: str = 'canine_output', dry_run: bool = False) -> pd.DataFrame:
        """
        Runs the configured pipeline
        Returns a pandas DataFrame containing job inputs, outputs, and runtime information
        """
        if isinstance(self.backend, LocalSlurmBackend) and os.path.exists(output_dir):
            raise FileExistsError("Output directory {} already exists".format(output_dir))
        job_spec = self.adapter.parse_inputs(self.raw_inputs)

        if len(job_spec) == 0:
            raise ValueError("You didn't specify any jobs!")
        elif len(job_spec) > 4000000:
            raise ValueError("Cannot exceede 4000000 jobs in one pipeline")

        print("Preparing pipeline of", len(job_spec), "jobs")
        print("Connecting to backend...")
        if isinstance(self.backend, RemoteSlurmBackend):
            self.backend.load_config_args()
        start_time = time.monotonic()
        with self.backend:
            print("Initializing pipeline workspace")
            with self._localizer_type(self.backend, **self.localizer_args) as localizer:
                print("Localizing inputs...")
                abs_staging_dir = localizer.localize(
                    job_spec,
                    self.raw_outputs,
                    self.localizer_overrides
                )
                print("Job staged on SLURM controller in:", abs_staging_dir)
                print("Preparing pipeline script")
                env = localizer.environment('compute')
                root_dir = env['CANINE_ROOT']
                entrypoint_path = os.path.join(root_dir, 'entrypoint.sh')
                if isinstance(self.script, str):
                    pipeline_path = os.path.join(root_dir, os.path.basename(self.script))
                else:
                    pipeline_path = self.backend.pack_batch_script(
                        *self.script,
                        script_path=os.path.join(root_dir, 'script.sh')
                    )
                with self.backend.transport() as transport:
                    if isinstance(self.script, str):
                        transport.send(self.script, pipeline_path)
                        transport.chmod(pipeline_path, 0o775)
                    with transport.open(entrypoint_path, 'w') as w:
                        w.write(ENTRYPOINT.format(
                            backend=self._backend_type,
                            adapter=self._adapter_type,
                            pipeline_script=pipeline_path,
                            **env
                        ))
                    transport.chmod(entrypoint_path, 0o775)
                    transport.chmod(pipeline_path, 0o775)
                if dry_run:
                    localizer.clean_on_exit = False
                    return job_spec
                print("Waiting for cluster to finish startup...")
                self.backend.wait_for_cluster_ready()

                # perform hard reset of cluster; some backends do this own their
                # own, in which case we skip.  we also can't do this if path to slurm.conf
                # is unknown.
                if self.backend.hard_reset_on_orch_init and self._slurmconf_path:
                    active_jobs = self.backend.squeue('all')
                    if len(active_jobs):
                        print("There are active jobs. Skipping slurmctld restart")
                    else:
                        try:
                            print("Stopping slurmctld")
                            rc, stdout, stderr = self.backend.invoke(
                                'sudo pkill slurmctld',
                                True
                            )
                            check_call('sudo pkill slurmctld', rc, stdout, stderr)
                            print("Loading configurations", self._slurmconf_path)
                            rc, stdout, stderr = self.backend.invoke(
                                'sudo slurmctld -c -f {}'.format(self._slurmconf_path),
                                True
                            )
                            check_call('sudo slurmctld -c -f {}'.format(self._slurmconf_path), rc, stdout, stderr)
                            print("Restarting slurmctl")
                            rc, stdout, stderr = self.backend.invoke(
                                'sudo slurmctld reconfigure',
                                True
                            )
                            check_call('sudo slurmctld reconfigure', rc, stdout, stderr)
                        except CalledProcessError:
                            traceback.print_exc()
                            print("Slurmctld restart failed")
                print("Submitting batch job")
                batch_id = self.backend.sbatch(
                    entrypoint_path,
                    **{
                        'requeue': True,
                        'job_name': self.name,
                        'array': "0-{}".format(len(job_spec)-1),
                        'output': "{}/%a/stdout".format(env['CANINE_JOBS']),
                        'error': "{}/%a/stderr".format(env['CANINE_JOBS']),
                        **self.resources
                    }
                )
                print("Batch id:", batch_id)
                completed_jobs = []
                cpu_time = {}
                uptime = {}
                prev_acct = None
                try:
                    waiting_jobs = {
                        '{}_{}'.format(batch_id, i)
                        for i in range(len(job_spec))
                    }
                    outputs = {}
                    while len(waiting_jobs):
                        time.sleep(30)
                        acct = self.backend.sacct(job=batch_id, format="JobId,State,ExitCode,CPUTimeRAW").astype({'CPUTimeRAW': int})
                        for jid in [*waiting_jobs]:
                            if jid in acct.index:
                                if prev_acct is not None and jid in prev_acct.index and prev_acct['CPUTimeRAW'][jid] > acct['CPUTimeRAW'][jid]:
                                    # Job has restarted since last update:
                                    if jid in cpu_time:
                                        cpu_time[jid] += prev_acct['CPUTimeRAW'][jid]
                                    else:
                                        cpu_time[jid] = prev_acct['CPUTimeRAW'][jid]
                                if acct['State'][jid] not in {'RUNNING', 'PENDING', 'NODE_FAIL'}:
                                    job = jid.split('_')[1]
                                    print("Job",job, "completed with status", acct['State'][jid], acct['ExitCode'][jid].split(':')[0])
                                    completed_jobs.append((job, jid))
                                    waiting_jobs.remove(jid)
                        for node in {node for node in self.backend.squeue(jobs=batch_id)['NODELIST(REASON)'] if not node.startswith('(')}:
                            if node in uptime:
                                uptime[node] += 1
                            else:
                                uptime[node] = 1
                        if prev_acct is None:
                            prev_acct = acct
                        else:
                            prev_acct = pd.concat([
                                acct,
                                prev_acct.loc[[idx for idx in prev_acct.index if idx not in acct.index]]
                            ])
                except:
                    print("Encountered unhandled exception. Cancelling batch job", file=sys.stderr)
                    self.backend.scancel(batch_id)
                    localizer.clean_on_exit = False
                    raise
                finally:
                    if len(completed_jobs):
                        print("Delocalizing outputs")
                        outputs = localizer.delocalize(self.raw_outputs, output_dir)
            print("Parsing output data")
            self.adapter.parse_outputs(outputs)
            acct = self.backend.sacct(job=batch_id)
        runtime = time.monotonic() - start_time
        df = pd.DataFrame(
            data={
                job_id: {
                    'slurm_state': acct['State'][batch_id+'_'+job_id],
                    'exit_code': acct['ExitCode'][batch_id+'_'+job_id],
                    'cpu_hours': (prev_acct['CPUTimeRAW'][batch_id+'_'+job_id] + (
                        cpu_time[batch_id+'_'+job_id] if batch_id+'_'+job_id in cpu_time else 0
                    ))/3600,
                    **job_spec[job_id],
                    **{
                        key: val[0] if isinstance(val, list) and len(val) == 1 else val
                        for key, val in outputs[job_id].items()
                    }
                }
                for job_id in job_spec
            }
        ).T.set_index(pd.Index([*job_spec], name='job_id')).astype({'cpu_hours': int})
        try:
            print("Estimated total cluster cost:", self.backend.estimate_cost(
                runtime/3600,
                node_uptime=sum(uptime.values())/120
            )[0])
            job_cost = self.backend.estimate_cost(job_cpu_time=df['cpu_hours'].to_dict())[1]
            df['est_cost'] = [job_cost[job_id] for job_id in df.index] if job_cost is not None else [0] * len(df)
        except:
            traceback.print_exc()
        finally:
            return df



# vim: set expandtab:

import typing
import subprocess
import os
import sys

from .local import LocalSlurmBackend
from ..utils import get_default_gcp_project, gcp_hourly_cost

import googleapiclient.discovery as gd
import pandas as pd

gce = gd.build('compute', 'v1')

def list_instances(zone: str, project: str) -> pd.DataFrame:
    """
    List all instances in a given zone and project
    """
    inst_dict = gce.instances().list(project = project, zone = zone).execute()

    fnames = ['name', 'machineType', 'status', 'zone', 'selfLink', 'tags']

    if "items" in inst_dict:
        inst_DF = pd.DataFrame([[x[y] for y in fnames] for x in inst_dict['items']], columns = fnames)
        inst_DF["tags"] = inst_DF["tags"].apply(lambda x : x["items"] if "items" in x else [])
    else:
        return pd.DataFrame()

    # API returns selfLinks; parse these into something human-readable
    return inst_DF.apply(lambda x : x.str.replace(r'.*/', '')
                         if x.name in ["machineType", "zone"] else x)

class TransientImageSlurmBackend(LocalSlurmBackend): # {{{
    """
    Backend for starting a Slurm cluster using a preconfigured GCE image.
    The image must meet the following requirements:
    * The current node is a valid Slurm controller, i.e.:
        * `slurmctld`, `munged`, and `slurmdbd` run properly (note that these do not need
          to already be running when invoking Canine with this backend; they will
          be started automatically as needed.)
        * Accounting is enabled (i.e., `sacct` can list completed jobs)
    * The default Slurm partition is compatible with any nodes that get spun up:
        * Worker node names must match names specified in a partition defined in `slurm.conf`
        * Worker node types must be consistent with node definitions in `slurm.conf`
    * The image provided has a valid Slurm installation, compatible with that of the
      controller node (e.g., same version, same plugins, etc.)
    * If GPUs are added, drivers must already be installed
    """

    def __init__(
        self, *, image: str, worker_prefix: str = 'slurm-canine', tot_node_count: int = 50,
        init_node_count: typing.Optional[int] = None, compute_zone: str = 'us-central1-a',
        worker_type: str = 'n1-highcpu-2', preemptible: bool = True,
        gpu_type: typing.Optional[str] = None, gpu_count: int = 0,
        compute_script_file: typing.Optional[str] = None,
        compute_script: typing.Optional[str] = None,
        project: typing.Optional[str] = None,
        user: typing.Optional[str] = None, slurm_conf_path: typing.Optional[str] = None,
        delete_on_stop: bool = False,
        **kwargs
    ):
        #
        # validate inputs that will not be caught later on (e.g. by gcloud invocations)
        if tot_node_count < 1:
            raise ValueError("tot_node_count cannot be less than 1.")

        if init_node_count is not None:
            if init_node_count < 0:
                raise ValueError("init_node_count cannot be negative.")

            if init_node_count > tot_node_count:
                raise ValueError("init_node_count cannot exceed tot_node_count.")

        if compute_script_file is not None and compute_script is not None:
            raise ValueError("Cannot simultaneously specifiy compute_script_file and compute_script.")

        if slurm_conf_path is None:
            raise ValueError("Currently, path to slurm.conf must be explicitly specified.")

        # make config dict
        self.config = {
            "image" : image,
            "worker_prefix" : worker_prefix,
            "tot_node_count" : tot_node_count,
            "init_node_count" : init_node_count if init_node_count else tot_node_count,
            "compute_zone" : compute_zone,
            "worker_type" : worker_type,
            "preemptible" : "--preemptible" if preemptible else "",
            "gpu_type" : gpu_type,
            "gpu_count" : gpu_count,
            "compute_script_file" :
                "--metadata-from-file startup-script={}".format(compute_script_file)
                if compute_script_file else "",
            "compute_script" :
                "--metadata startup-script={}".format(compute_script)
                if compute_script else "",
            "project" : project if project else get_default_gcp_project(),
            "user" : user if user else "root",
            "slurm_conf_path" : slurm_conf_path,
            "delete_on_stop" : delete_on_stop
        }

        # this backend resets itself on startup; no need for the orchestrator
        # to do this.
        self.hard_reset_on_orch_init = False

        # list of nodes under the purview of Canine
        self.nodes = pd.DataFrame()

        # export slurm_conf_path to environment
        os.environ['SLURM_CONF'] = self.config["slurm_conf_path"]

    def __enter__(self):
        try:
            #
            # check if Slurm is already running locally; start (with hard reset) if not

            print("Checking for running Slurm controller ... ", end = "", flush = True)

            subprocess.check_call(
                """sudo -E -u {user} bash -c 'pgrep slurmctld || slurmctld -c -f {slurm_conf_path} &&
                   slurmctld reconfigure; pgrep slurmdbd || slurmdbd; pgrep munged || munged -f'
                """.format(**self.config),
                shell = True,
                stdout = subprocess.DEVNULL
            )

            # ensure all started successfully
            subprocess.check_call("pgrep slurmctld && pgrep slurmdbd && pgrep munged", shell = True,
                                  stdout = subprocess.DEVNULL)

            print("done", flush = True)

            #
            # create/start worker nodes

            print("Checking for preexisting cluster nodes ... ", end = "", flush = True)

            nodenames = pd.DataFrame(index = [
                          self.config["worker_prefix"] + str(x) for x in
                          range(1, self.config["tot_node_count"] + 1)
                        ])

            # check which worker nodes exist outside of Canine
            instances = self.list_instances_all_zones()
            k9_inst_idx = instances["tags"].apply(lambda x : "caninetransientimage" in x)

            nodenames["is_ex_node"] = nodenames.index.isin(instances.loc[~k9_inst_idx, "name"])

            # check which worker nodes were previous created by Canine
            nodenames["is_k9_node"] = nodenames.index.isin(instances.loc[k9_inst_idx, "name"])

            instances = instances.merge(nodenames, left_on = "name", right_index = True,
                          how = "right")

            print("done", flush = True)

            # handle nodes that will not be created
            if (nodenames["is_ex_node"] | nodenames["is_k9_node"]).any():
                # WARN about worker nodes outside of Canine
                ex_nodes = nodenames.index[nodenames["is_ex_node"]]

                if not ex_nodes.empty:
                    print("WARNING: the following nodes already exist outside of Canine:\n - ",
                      file = sys.stderr, end = "")
                    print("\n - ".join(ex_nodes), file = sys.stderr)
                    print("Canine will not touch these nodes.")

                # ERROR if any instance types (Canine or external) are incongruous
                # with given definition
                typemismatch_idx = ~instances.loc[:, "machineType"].isna() & \
                                   (instances.loc[:, "machineType"] != self.config["worker_type"])

                if typemismatch_idx.any():
                    print("ERROR: nodes that already exist do not match specified machine type ({worker_type}):".format(**self.config), file = sys.stderr)
                    print(instances.drop(columns = "selfLink").loc[typemismatch_idx] \
                      .to_string(index = False), file = sys.stderr)
                    print("This will result in Slurm bitmap corruption.", file = sys.stderr)

                    raise RuntimeError('Preexisting cluster nodes do not match specified machine type for new nodes')

                # WARN if any nodes (Canine or external) are already defined but
                # present in other zones
                zonemismatch_idx = ~instances.loc[:, "zone"].isna() & \
                                   (instances.loc[:, "zone"] != self.config["compute_zone"])

                if zonemismatch_idx.any():
                    print("WARNING: nodes that already exist do not match specified compute zone ({compute_zone}):".format(**self.config), file = sys.stderr)
                    print(instances.drop(columns = "selfLink").loc[zonemismatch_idx] \
                      .to_string(index = False), file = sys.stderr)
                    print("This may result in degraded performance or egress charges.",
                      file = sys.stderr)

            self.nodes = nodenames.loc[~nodenames["is_ex_node"]]

            # create the nodes

            # TODO: support the other config flags
            # TODO: use API to launch these

            if (~self.nodes["is_k9_node"]).any():
                nodes_to_create = self.nodes.index[~self.nodes["is_k9_node"]].values

                print("Creating {0:d} worker nodes ... ".format(nodes_to_create.shape[0]),
                      end = "", flush = True)
                subprocess.check_call(
                    """gcloud compute instances create {workers} \
                       --image {image} --machine-type {worker_type} --zone {compute_zone} \
                       {compute_script} {compute_script_file} {preemptible} \
                       --tags caninetransientimage
                    """.format(**self.config, workers = " ".join(nodes_to_create)),
                    shell = True
                )
                print("done", flush = True)

            # start nodes previously created by Canine

            instances_to_start = instances.loc[instances["is_k9_node"] &
              (instances["status"] == "TERMINATED"), "name"]

            # TODO: use API to start these
            if instances_to_start.shape[0] > 0:
                print("Starting {0:d} preexisting worker nodes ... ".format(instances_to_start.shape[0]),
                      end = "", flush = True)
                subprocess.check_call(
                    """gcloud compute instances start {workers} --zone {compute_zone} \
                    """.format(**self.config, workers = " ".join(instances_to_start.values)),
                    shell = True
                )
                print("done", flush = True)

            #
            # shut down nodes exceeding init_node_count

            for node in self.nodes.index[self.config["init_node_count"]:]:
                try:
                    self._pzw(gce.instances().stop)(instance = node).execute()
                except Exception as e:
                    print("WARNING: couldn't shutdown instance {}".format(node), file = sys.stderr)
                    print(e)

            return self
        except Exception as e:
            print("ERROR: Could not initialize cluster; attempting to tear down.", file = sys.stderr)

            self.stop()
            raise e

    def __exit__(self, *args):
        self.stop()

    def stop(self, delete_on_stop = None):
        """
        Delete or stop (default) compute instances
        """
        if delete_on_stop is None:
            delete_on_stop = self.config["delete_on_stop"]

        #
        # stop or delete compute nodes

        for node in self.nodes.index:
            try:
                if delete_on_stop:
                    self._pzw(gce.instances().delete)(instance = node).execute()
                else:
                    self._pzw(gce.instances().stop)(instance = node).execute()
            except Exception as e:
                print("WARNING: couldn't shutdown instance {}".format(node), file = sys.stderr)
                print(e)

    def list_instances_all_zones(self):
        """
        List all instances across all zones in `self.config["project"]`
        """
        zone_dict = gce.zones().list(project = self.config["project"]).execute()

        return pd.concat([
          list_instances(zone = x["name"], project = self.config["project"])
          for x in zone_dict["items"]
        ], axis = 0).reset_index(drop = True)

    # a handy wrapper to automatically add this instance's project and zone to
    # GCP API calls
    # TODO: for bonus points, can we automatically apply this to all relevant
    #       methods in a GCE class instance?
    def _pzw(self, f):
        def x(*args, **kwargs):
            return f(project = self.config["project"], zone = self.config["compute_zone"], *args, **kwargs)

        return x

    def estimate_cost(self, clock_uptime: typing.Optional[float] = None, node_uptime: typing.Optional[float] = None, job_cpu_time: typing.Optional[typing.Dict[str, float]] = None) -> typing.Tuple[float, typing.Optional[typing.Dict[str, float]]]:
        """
        Returns a cost estimate for the cluster, based on any cost information available
        to the backend. May provide total node uptime (for cluster cost estimate)
        and/or cpu_time for each job to get job specific cost estimates.
        Clock uptime may be provided and is useful if the cluster has an inherrant
        overhead for uptime (ie: controller nodes).
        Note: Job cost estimates may not sum up to the total cluster cost if the
        cluster was not at full utilization.
        """
        cluster_cost = 0
        worker_cpu_cost = 0
        job_cost = None
        if node_uptime is not None:
            worker_info = {
                'mtype': self.config['worker_type'],
                'preemptible': bool(self.config['preemptible'])
            }
            if 'gpu_type' in self.config and 'gpu_count' in self.config and self.config['gpu_count'] > 0:
                worker_info['gpu_type'] = self.config['gpu_type']
                worker_info['gpu_count'] = self.config['gpu_count']
            worker_hourly_cost = gcp_hourly_cost(**worker_info)
            cluster_cost = node_uptime * worker_hourly_cost
            mtype_prefix = self.config['worker_type'][:3]
            if mtype_prefix in {'f1-', 'g1-'}:
                ncpus = 1
            elif mtype_prefix == 'cus': # n1-custom-X
                ncpus = int(self.config['worker_type'].split('-')[1])
            else:
                ncpus = int(self.config['worker_type'].split('-')[2])
            # Approximates the cost burden / CPU hour of the VM
            worker_cpu_cost = worker_hourly_cost / ncpus
        if job_cpu_time is not None:
            job_cost = {
                job_id: worker_cpu_cost * cpu_time
                for job_id, cpu_time in job_cpu_time.items()
            }
        return cluster_cost, job_cost



# }}}



import typing
import os
import time
import sys
import warnings
import traceback
import shlex
import tempfile
from subprocess import CalledProcessError
from .backends import AbstractSlurmBackend, LocalSlurmBackend
from .utils import check_call
from .orchestrator import Orchestrator, BACKENDS, version
import yaml
import pandas as pd
from agutil import status_bar

class Xargs(Orchestrator):
    """
    Simplified Orchestrator
    Built to work on a local slurm cluster.
    No localization is performed. Job-specific commands are dumped to a staging
    directory, and jobs are dispatched. Outputs are not gathered or delocalized,
    and the only cleanup performed removes the temporary staging dir
    """

    def __init__(
        self, command: str, inputs: typing.Dict[str, typing.Any],
        backend: typing.Union[AbstractSlurmBackend, typing.Dict[str, typing.Any]],
        name: typing.Optional[str] = None,
        cwd: typing.Optional[str] = None,
        resources: typing.Optional[typing.Dict[str, typing.Any]] = None
    ):
        if isinstance(backend, AbstractSlurmBackend):
            self.backend = backend
            self._slurmconf_path = None
        else:
            self.backend = BACKENDS[backend['type']](**backend)
            self._slurmconf_path = backend['slurm_conf_path'] if 'slurm_conf_path' in backend else None
        if not isinstance(self.backend, LocalSlurmBackend):
            raise TypeError("Xargs only works on local-based slurm backends")
        self.resources = {} if resources is None else resources
        self.inputs = inputs
        self.cwd = cwd
        self.n_jobs = 0 if len(inputs) == 0 else len(inputs['canine_arg0'])
        self.name = name if name is not None else 'Canine-Xargs'
        self.command = command

    def run_pipeline(self, dry_run: bool = False) -> pd.DataFrame:
        print("Preparing a pipeline of", self.n_jobs, "jobs")
        print("Connecting to backend...")
        start_time = time.monotonic()
        with self.backend:
            print("Initializing pipeline workspace")
            with tempfile.TemporaryDirectory() as tempdir:
                for jid in range(self.n_jobs):
                    # By creating a local tempdir
                    # but using the backend to write the script
                    # we ensure that this is a local-based backend
                    self.backend.pack_batch_script(
                        *(
                            'export {}={}'.format(
                                argname,
                                shlex.quote(argvalues[jid])
                            )
                            for argname, argvalues in self.inputs.items()
                        ),
                        script_path=os.path.join(
                            tempdir,
                            '{}.sh'.format(jid)
                        )
                    )
                print("Job staged on SLURM controller in:", tempdir)
                print("Preparing pipeline script")
                self.backend.pack_batch_script(
                    'export CANINE={}'.format(version),
                    'export CANINE_BACKEND='.format(type(self.backend)),
                    'export CANINE_ADAPTER=Xargs',
                    'export CANINE_ROOT={}'.format(tempdir),
                    'export CANINE_COMMON=""',
                    'export CANINE_OUTPUT=""',
                    'export CANINE_JOBS={}'.format(self.cwd),
                    'source $CANINE_ROOT/$SLURM_ARRAY_TASK_ID.sh',
                    self.command,
                    script_path=os.path.join(
                        tempdir,
                        'entrypoint.sh'
                    )
                )
                if dry_run:
                    return
                print("Waiting for cluster to finish startup...")
                self.backend.wait_for_cluster_ready()
                if self.backend.hard_reset_on_orch_init and self._slurmconf_path:
                    active_jobs = self.backend.squeue('all')
                    if len(active_jobs):
                        print("There are active jobs. Skipping slurmctld restart")
                    else:
                        try:
                            print("Stopping slurmctld")
                            rc, stdout, stderr = self.backend.invoke(
                                'sudo pkill slurmctld',
                                True
                            )
                            check_call('sudo pkill slurmctld', rc, stdout, stderr)
                            print("Loading configurations", self._slurmconf_path)
                            rc, stdout, stderr = self.backend.invoke(
                                'sudo slurmctld -c -f {}'.format(self._slurmconf_path),
                                True
                            )
                            check_call('sudo slurmctld -c -f {}'.format(self._slurmconf_path), rc, stdout, stderr)
                            print("Restarting slurmctl")
                            rc, stdout, stderr = self.backend.invoke(
                                'sudo slurmctld reconfigure',
                                True
                            )
                            check_call('sudo slurmctld reconfigure', rc, stdout, stderr)
                        except CalledProcessError:
                            traceback.print_exc()
                            print("Slurmctld restart failed")
                print("Submitting batch job")
                batch_id = self.backend.sbatch(
                    os.path.join(
                        tempdir,
                        'entrypoint.sh'
                    ),
                    **{
                        'chdir': '~' if self.cwd is None else self.cwd,
                        'requeue': True,
                        'job_name': self.name,
                        'array': "0-{}".format(self.n_jobs-1),
                        'output': "{}/%A.%a.stdout".format('~' if self.cwd is None else self.cwd),
                        'error': "{}/%A.%a.stderr".format('~' if self.cwd is None else self.cwd),
                        **self.resources
                    }
                )
                print("Batch id:", batch_id)
                completed_jobs = []
                cpu_time = {}
                uptime = {}
                prev_acct = None
                try:
                    waiting_jobs = {
                        '{}_{}'.format(batch_id, i)
                        for i in range(self.n_jobs)
                    }
                    while len(waiting_jobs):
                        time.sleep(30)
                        acct = self.backend.sacct(job=batch_id, format="JobId,State,ExitCode,CPUTimeRAW").astype({'CPUTimeRAW': int})
                        for jid in [*waiting_jobs]:
                            if jid in acct.index:
                                if prev_acct is not None and jid in prev_acct.index and prev_acct['CPUTimeRAW'][jid] > acct['CPUTimeRAW'][jid]:
                                    # Job has restarted since last update:
                                    if jid in cpu_time:
                                        cpu_time[jid] += prev_acct['CPUTimeRAW'][jid]
                                    else:
                                        cpu_time[jid] = prev_acct['CPUTimeRAW'][jid]
                                if acct['State'][jid] not in {'RUNNING', 'PENDING', 'NODE_FAIL'}:
                                    job = jid.split('_')[1]
                                    print("Job",job, "completed with status", acct['State'][jid], acct['ExitCode'][jid].split(':')[0])
                                    completed_jobs.append((job, jid))
                                    waiting_jobs.remove(jid)
                        for node in {node for node in self.backend.squeue(jobs=batch_id)['NODELIST(REASON)'] if not node.startswith('(')}:
                            if node in uptime:
                                uptime[node] += 1
                            else:
                                uptime[node] = 1
                        if prev_acct is None:
                            prev_acct = acct
                        else:
                            prev_acct = pd.concat([
                                acct,
                                prev_acct.loc[[idx for idx in prev_acct.index if idx not in acct.index]]
                            ])
                except:
                    print("Encountered unhandled exception. Cancelling batch job", file=sys.stderr)
                    self.backend.scancel(batch_id)
                    raise
        runtime = time.monotonic() - start_time
        job_spec = {
            str(i): {
                argname: argvalues[i]
                for argname, argvalues in self.inputs.items()
            }
            for i in range(self.n_jobs)
        }
        df = pd.DataFrame(
            data={
                job_id: {
                    'slurm_state': acct['State'][batch_id+'_'+job_id],
                    'exit_code': acct['ExitCode'][batch_id+'_'+job_id],
                    'cpu_hours': (prev_acct['CPUTimeRAW'][batch_id+'_'+job_id] + (
                        cpu_time[batch_id+'_'+job_id] if batch_id+'_'+job_id in cpu_time else 0
                    ))/3600,
                    **job_spec[job_id],
                }
                for job_id in job_spec
            }
        ).T.set_index(pd.Index([*job_spec], name='job_id')).astype({'cpu_hours': int})
        try:
            print("Estimated total cluster cost:", self.backend.estimate_cost(
                runtime/3600,
                node_uptime=sum(uptime.values())/120
            )[0])
            job_cost = self.backend.estimate_cost(job_cpu_time=df['cpu_hours'].to_dict())[1]
            df['est_cost'] = [job_cost[job_id] for job_id in df.index] if job_cost is not None else [0] * len(df)
        except:
            traceback.print_exc()
        finally:
            return df



"""
canine.backends
=================================
Contains SLURM backends for canine
"""

from .base import AbstractTransport, AbstractSlurmBackend
from .local import LocalTransport, LocalSlurmBackend
from .remote import RemoteTransport, RemoteSlurmBackend
from .gcpTransient import TransientGCPSlurmBackend
from .imageTransient import TransientImageSlurmBackend

__all__ = [
    'LocalSlurmBackend',
    'LocalTransport',
    'RemoteSlurmBackend',
    'RemoteTransport',
    'TransientGCPSlurmBackend',
    'TransientImageSlurmBackend'
]



import getpass
import tempfile
import typing
import time
import subprocess
import shutil
import os
import sys
from .remote import RemoteSlurmBackend
from ..utils import get_default_gcp_project, ArgumentHelper, check_call, gcp_hourly_cost
# import paramiko
import yaml
import pandas as pd

SLURM_PARTITION_RECON = b'slurm_load_partitions: Unable to contact slurm controller (connect failure)'
PARAMIKO_PEM_KEY = os.path.expanduser('~/.ssh/canine_pem_key')
GPU_SCRIPT = ' && '.join([
    # 'sudo yum install -y kernel-devel-$(uname -r) kernel-headers-$(uname -r)',
    # 'wget http://developer.download.nvidia.com/compute/cuda/repos/rhel7/x86_64/cuda-repo-rhel7-10.1.168-1.x86_64.rpm',
    # 'sudo yum install -y cuda-repo-rhel7-10.1.168-1.x86_64.rpm',
    # 'sudo yum -y updateinfo',
    # 'sudo yum install -y cuda',
    # 'wget http://developer.download.nvidia.com/compute/machine-learning/repos/rhel7/x86_64/nvidia-machine-learning-repo-rhel7-1.0.0-1.x86_64.rpm',
    # 'sudo yum install -y nvidia-machine-learning-repo-rhel7-1.0.0-1.x86_64.rpm',
    # 'sudo yum -y updateinfo',
    # 'sudo yum install -y cuda-10-0 libcudnn7 libcudnn7-devel libnvinfer5',
    # 'curl -s -L https://nvidia.github.io/nvidia-docker/$(. /etc/os-release;echo $ID$VERSION_ID)/nvidia-docker.repo | sudo tee /etc/yum.repos.d/nvidia-docker.repo',
    # 'sudo yum -y updateinfo',
    # 'sudo yum install -y nvidia-docker2'
    'curl -s -L https://nvidia.github.io/nvidia-docker/gpgkey | sudo apt-key add -',
    'curl -s -L https://nvidia.github.io/nvidia-docker/$(. /etc/os-release;echo $ID$VERSION_ID)/nvidia-docker.list | sudo tee /etc/apt/sources.list.d/nvidia-docker.list',
    'sudo apt-get update',
    'sudo apt-get install -y nvidia-container-toolkit nvidia-docker2'
])
GPU_TYPES = {
    'nvidia-tesla-k80',
    'nvidia-tesla-p100',
    'nvidia-tesla-v100',
    'nvidia-tesla-p4',
    'nvidia-tesla-t4'
}

class TransientGCPSlurmBackend(RemoteSlurmBackend):
    """
    Backend for transient slurm clusters which need to be deployed and configured
    on GCP before they're ready for use
    """

    def __init__(
        self, name: str = 'slurm-canine', *, max_node_count: int = 10, compute_zone: str = 'us-central1-a',
        controller_type: str = 'n1-standard-16', login_type: str = 'n1-standard-1', preemptible: bool = True,
        worker_type: str = 'n1-highcpu-2', login_count: int = 0, compute_disk_size: int = 20,
        controller_disk_size: int = 200, gpu_type: typing.Optional[str] = None, gpu_count: int = 0,
        compute_script: str = "", controller_script: str = "", secondary_disk_size: int = 0, project: typing.Optional[str]  = None,
        **kwargs : typing.Any
    ):
        self.project = project if project is not None else get_default_gcp_project()
        super().__init__('{}-controller.{}.{}'.format(
            name,
            compute_zone,
            self.project
        ))
        self.config = {
          "cluster_name": name,
          "static_node_count": 0,
          "max_node_count": int(max_node_count),
          "zone": compute_zone,
          "region": compute_zone[:-2],
          "cidr": "10.10.0.0/16",
          "controller_machine_type": controller_type,
          "compute_machine_type": worker_type,
          "compute_disk_type": "pd-standard",
          "compute_disk_size_gb": int(compute_disk_size),
          "controller_disk_type": "pd-ssd",
          "controller_disk_size_gb": int(controller_disk_size),
          "controller_secondary_disk": secondary_disk_size > 0,
          "login_machine_type": login_type,
          "login_node_count": int(login_count),
          "login_disk_size_gb": 10,
          "preemptible_bursting": preemptible,
          "private_google_access": True,
          "vpc_net": "default",
          "vpc_subnet": "default",
          "default_users": getpass.getuser(),
          'gpu_count': 0,
          'slurm_version': '19.05-latest',
          **kwargs
        }

        if gpu_type is not None and gpu_count > 0:
            if gpu_type not in GPU_TYPES:
                raise ValueError("gpu_type must be one of {}".format(GPU_TYPES))
            self.config['gpu_type'] = gpu_type
            self.config['gpu_count'] = gpu_count

        if secondary_disk_size > 0:
            self.config['controller_secondary_disk_type'] = 'pd-standard'
            self.config['controller_secondary_disk_size_gb'] = secondary_disk_size

        self.startup_script = """
        sudo apt-get install -y apt-utils lvm2 cgroup-bin cgroup-tools libcgroup-dev htop gcc python-dev python-setuptools wget docker.io
        sudo groupadd docker
        sudo usermod -aG docker {0}
        sudo usermod -aG sudo {0}
        sudo systemctl enable docker.service
        sudo systemctl start docker.service
        sudo chown root:docker /var/run/docker.sock
        sudo bash -c "echo {0}$'\\t'ALL='(ALL:ALL)'$'\\t'NOPASSWD:$'\\t'ALL >> /etc/sudoers"
        sudo sed -e 's/GRUB_CMDLINE_LINUX="\\?\\([^"]*\\)"\\?/GRUB_CMDLINE_LINUX="\\1 cgroup_enable=memory swapaccount=1"/' < /etc/default/grub > grub.tmp
        sudo mv grub.tmp /etc/default/grub
        sudo grub2-mkconfig -o /etc/grub2.cfg
        curl https://bootstrap.pypa.io/get-pip.py -o get-pip.py
        sudo python get-pip.py
        sudo pip uninstall -y crcmod
        sudo pip install --no-cache-dir -U crcmod
        {1}
        {2}
        """.format(
            getpass.getuser(),
            GPU_SCRIPT if self.config['gpu_count'] > 0 else '',
            compute_script
        )

        self.controller_script = """
        sudo dd if=/dev/zero of=/swapfile count=4096 bs=1MiB
        sudo chmod 700 /swapfile
        sudo mkswap /swapfile
        sudo swapon /swapfile
        sudo apt-get install -y gcc python-dev python-setuptools htop
        curl https://bootstrap.pypa.io/get-pip.py -o get-pip.py
        sudo python get-pip.py
        sudo pip uninstall -y crcmod
        sudo pip install --no-cache-dir -U crcmod
        {0}
        """.format(
            controller_script
        )

        subprocess.check_call(
            'touch ~/.ssh/google_compute_known_hosts',
            shell=True,
            executable='/bin/bash'
        )

    def __enter__(self):
        """
        Create NFS server
        Set up the NFS server and SLURM cluster
        Use default VPC
        """
        # Idea: Allow the final deployment step to run in a Popen
        # cluster.run then polls the object to make sure the deployment has finished
        # allows copytree to begin while deployment still in progress

        # User may also provide an NFS server IP

        # 1) create NFS: (n1-highcpu-16 [16CPU, 14GB, 4GB SW], SSD w/ user GB)
        # 2) SCP setup script to NFS
        # 3) SSH into NFS, run script
        try:
            with tempfile.TemporaryDirectory() as tempdir:
                shutil.copytree(
                    os.path.join(
                        os.path.dirname(__file__),
                        'slurm-gcp'
                    ),
                    tempdir+'/slurm'
                )
                with open(os.path.join(tempdir, 'slurm', 'scripts', 'custom-compute-install'), 'w') as w:
                    w.write(self.startup_script)
                with open(os.path.join(tempdir, 'slurm', 'scripts', 'custom-controller-install'), 'w') as w:
                    w.write(self.controller_script)
                with open(os.path.join(tempdir, 'slurm', 'slurm-cluster.yaml'), 'w') as w:
                    yaml.dump(
                        {
                            "imports": [
                              {
                                "path": 'slurm.jinja'
                              }
                            ],
                            "resources": [
                              {
                                "name": "slurm-cluster",
                                "type": "slurm.jinja",
                                "properties": self.config
                              }
                            ]
                        },
                        w
                    )
                subprocess.check_call(
                    'gcloud deployment-manager deployments create {} --config {} --project {}'.format(
                        self.config['cluster_name'],
                        os.path.join(tempdir, 'slurm', 'slurm-cluster.yaml'),
                        self.project
                    ),
                    shell=True,
                    executable='/bin/bash'
                )
            subprocess.check_call(
                'gcloud compute config-ssh --project {}'.format(self.project),
                shell=True
            )
            subprocess.check_call(
                'gcloud compute os-login ssh-keys add --key-file '
                '~/.ssh/google_compute_engine.pub --project {} >/dev/null'.format(
                    self.project
                ),
                shell=True
            )
            self.load_config_args()
            time.sleep(30) # Key propagation time
            super().__enter__()
            print("Waiting for slurm to initialize")
            rc, sout, serr = self.invoke("which sinfo")
            while rc != 0:
                time.sleep(10)
                rc, sout, serr = self.invoke("which sinfo")
            time.sleep(60)
            print("Slurm controller is ready. Please call .wait_for_cluster_ready() to wait until the slurm compute nodes are ready to accept work")
            return self
        except:
            self.stop()
            raise

    def stop(self):
        """
        Kills the slurm cluster
        """
        subprocess.check_call(
            'gcloud compute config-ssh --remove',
            shell=True
        )
        subprocess.check_call(
            'gcloud deployment-manager deployments delete {} --project {} -q'.format(
                self.config['cluster_name'],
                self.project
            ),
            shell=True,
            executable='/bin/bash'
        )
        subprocess.run(
            "gcloud compute images delete --project {0} --quiet "
            "$(gcloud compute images list --project {0} --filter family:{1}-compute-image-family| awk 'NR>1 {{print $1}}')".format(
                self.project,
                self.config['cluster_name']
            ),
            shell=True,
            executable='/bin/bash'
        )
        subprocess.run(
            "gcloud compute instances delete --project {0} --quiet "
            "--zone {1} $(gcloud compute instances list --project {0} --filter name:{2}-compute | awk 'NR>1 {{print $1}}')".format(
                self.project,
                self.config['zone'],
                self.config['cluster_name']
            ),
            shell=True,
            executable='/bin/bash'
        )

    def __exit__(self, *args):
        """
        kill NFS server
        delete the deployment
        """
        super().__exit__()
        self.stop()

    def sinfo(self, *slurmopts: str, **slurmparams: typing.Any) -> pd.DataFrame:
        """
        Shows the current cluster information
        slurmopts and slurmparams are passed into an ArgumentHelper and unpacked
        as command line arguments
        """
        command = 'sinfo'+ArgumentHelper(*slurmopts, **slurmparams).commandline
        status, stdout, stderr = self.invoke(command)
        if status != 0 and SLURM_PARTITION_RECON in stderr.read():
            print("Transient controller timed out while checking partitions. Retrying...", file=sys.stderr)
            time.sleep(120)
            status, stdout, stderr = self.invoke(command)
            if status != 0 and SLURM_PARTITION_RECON in stderr.read():
                print("Transient controller timed out while checking partitions. Making one final retry", file=sys.stderr)
                time.sleep(120)
                status, stdout, stderr = self.invoke(command)
        stderr.seek(0,0)
        check_call(command, status, stdout, stderr)
        df = pd.read_fwf(
            stdout,
            index_col=0
        )
        df.index = df.index.map(str)
        return df

    def wait_for_cluster_ready(self):
        """
        Blocks until the main partition is marked as up. Overrides base method,
        which waits until >= 1 node is in a ready state. Because TransientGCP
        starts all nodes in a powersave state, the base method would hang
        indefinitely.
        """
        df = self.sinfo()
        default = [value for value in df.index if value.endswith('*')]
        while len(default) == 0:
            time.sleep(10)
            df = self.sinfo()
            default = [value for value in df.index if value.endswith('*')]
        while (df.AVAIL[default] != 'up').all():
            time.sleep(10)
            df = self.sinfo()

    def estimate_cost(self, clock_uptime: typing.Optional[float] = None, node_uptime: typing.Optional[float] = None, job_cpu_time: typing.Optional[typing.Dict[str, float]] = None) -> typing.Tuple[float, typing.Optional[typing.Dict[str, float]]]:
        """
        Returns a cost estimate for the cluster, based on any cost information available
        to the backend. May provide total node uptime (for cluster cost estimate)
        and/or cpu_time for each job to get job specific cost estimates.
        Clock uptime may be provided and is useful if the cluster has an inherrant
        overhead for uptime (ie: controller nodes).
        Note: Job cost estimates may not sum up to the total cluster cost if the
        cluster was not at full utilization.
        """
        cluster_cost = 0
        worker_cpu_cost = 0
        job_cost = None
        if node_uptime is not None:
            worker_info = {
                'mtype': self.config['compute_machine_type'],
                'preemptible': self.config['preemptible_bursting']
            }
            if self.config['compute_disk_type'] == 'pd-ssd':
                worker_info['ssd_size'] = self.config['compute_disk_size_gb']
            else:
                worker_info['hdd_size'] = self.config['compute_disk_size_gb']
            if 'gpu_type' in self.config and 'gpu_count' in self.config and self.config['gpu_count'] > 0:
                worker_info['gpu_type'] = self.config['gpu_type']
                worker_info['gpu_count'] = self.config['gpu_count']
            worker_hourly_cost = gcp_hourly_cost(**worker_info)
            cluster_cost += node_uptime * worker_hourly_cost
            mtype_prefix = self.config['compute_machine_type'][:3]
            if mtype_prefix in {'f1-', 'g1-'}:
                ncpus = 1
            elif mtype_prefix == 'cus': # n1-custom-X
                ncpus = int(self.config['compute_machine_type'].split('-')[1])
            else:
                ncpus = int(self.config['compute_machine_type'].split('-')[2])
            # Approximates the cost burden / CPU hour of the VM
            worker_cpu_cost = worker_hourly_cost / ncpus
        if clock_uptime is not None:
            controller_info = {
                'mtype': self.config['controller_machine_type'],
                'preemptible': False,
            }
            if self.config['controller_disk_type'] == 'pd-ssd':
                controller_info['ssd_size'] = self.config['controller_disk_size_gb']
            else:
                controller_info['hdd_size'] = self.config['controller_disk_size_gb']
            if 'controller_secondary_disk_size_gb' in self.config:
                if 'hdd_size' in controller_info:
                    controller_info['hdd_size'] += self.config['controller_secondary_disk_size_gb']
                elif self.config['controller_secondary_disk_size_gb'] > 0:
                    controller_info['hdd_size'] = self.config['controller_secondary_disk_size_gb']
            cluster_cost += clock_uptime * gcp_hourly_cost(**controller_info)
        if job_cpu_time is not None:
            job_cost = {
                job_id: worker_cpu_cost * cpu_time
                for job_id, cpu_time in job_cpu_time.items()
            }
        return cluster_cost, job_cost



import typing
import warnings
import shutil
import tempfile
import os
import shlex
import fnmatch
from .base import PathType, Localization
from .local import BatchedLocalizer
from ..backends import AbstractSlurmBackend, AbstractTransport
from ..utils import get_default_gcp_project
from agutil import status_bar


class NFSLocalizer(BatchedLocalizer):
    """
    Similar to LocalLocalizer:
    Constructs the canine staging directory locally.
    Local inputs are symlinked into the staging directory.

    EXCEPT:
    Unlike LocalLocalizer, there is no final transfer step.
    Files are staged locally and this strategy expects that the given local staging dir
    is NFS mounted to the slurm cluster
    """

    def __init__(
        self, backend: AbstractSlurmBackend, transfer_bucket: typing.Optional[str] = None,
        common: bool = True, staging_dir: str = None, mount_path: str = None,
        project: typing.Optional[str] = None,
    ):
        """
        Initializes the Localizer using the given transport.
        Localizer assumes that the SLURMBackend is connected and functional during
        the localizer's entire life cycle.
        Note: staging_dir refers to the directory on the LOCAL filesystem
        Note: mount_path refers to the mounted directory on the REMOTE filesystem (both the controller and worker nodes)
        """
        if staging_dir is None:
            raise TypeError("staging_dir is a required argument for NFSLocalizer")
        if transfer_bucket is not None:
            warnings.warn("transfer_bucket has no bearing on NFSLocalizer. It is kept purely for adherence to the API")
        self.backend = backend
        self.common = common
        self.common_inputs = set()
        self.__sbcast = False
        if staging_dir == 'SBCAST':
            # FIXME: This doesn't actually do anything yet
            # If sbcast is true, then localization needs to use backend.sbcast to move files to the remote system
            # Not sure at all how delocalization would work
            self.__sbcast = True
            staging_dir = None
        self._local_dir = tempfile.TemporaryDirectory()
        self.local_dir = staging_dir
        if not os.path.isdir(self.local_dir):
            os.makedirs(self.local_dir)
        with self.backend.transport() as transport:
            self.mount_path = transport.normpath(mount_path if mount_path is not None else staging_dir)
        self.staging_dir = self.mount_path
        self.inputs = {} # {jobId: {inputName: (handle type, handle value)}}
        self.clean_on_exit = True
        self.project = project if project is not None else get_default_gcp_project()

    def localize_file(self, src: str, dest: PathType, transport: typing.Optional[AbstractTransport] = None):
        """
        Localizes the given file.
        gs:// files are copied to the staging directory
        local files are symlinked to the staging directory
        """
        if src.startswith('gs://'):
            self.gs_copy(
                src,
                dest.localpath,
                'local'
            )
        elif os.path.exists(src):
            src = os.path.abspath(src)
            if not os.path.isdir(os.path.dirname(dest.localpath)):
                os.makedirs(os.path.dirname(dest.localpath))
            if os.path.isfile(src):
                shutil.copyfile(src, dest.localpath)
            else:
                shutil.copytree(src, dest.localpath)

    def localize(self, inputs: typing.Dict[str, typing.Dict[str, str]], patterns: typing.Dict[str, str], overrides: typing.Optional[typing.Dict[str, typing.Optional[str]]] = None) -> str:
        """
        3 phase task:
        1) Pre-scan inputs to determine proper localization strategy for all inputs
        2) Begin localizing job inputs. For each job, check the predetermined strategy
        and set up the job's setup and teardown scripts
        3) Finally, finalize the localization. This may include broadcasting the
        staging directory or copying a batch of gsutil files
        Returns the remote staging directory, which is now ready for final startup
        """
        if overrides is None:
            overrides = {}
        overrides = {k:v.lower() if isinstance(v, str) else None for k,v in overrides.items()}
        with self.backend.transport() as transport:
            if self.common:
                common_dests = self.pick_common_inputs(inputs, overrides, transport=transport)
            else:
                common_dests = {}
            for jobId, data in inputs.items():
                os.makedirs(os.path.join(
                    self.environment('local')['CANINE_JOBS'],
                    jobId,
                ))
                self.prepare_job_inputs(jobId, data, common_dests, overrides, transport=transport)
                # Now localize job setup and teardown scripts
                setup_script, teardown_script = self.job_setup_teardown(jobId, patterns)
                # Setup:
                script_path = self.reserve_path('jobs', jobId, 'setup.sh')
                with open(script_path.localpath, 'w') as w:
                    w.write(setup_script)
                os.chmod(script_path.localpath, 0o775)
                # Teardown:
                script_path = self.reserve_path('jobs', jobId, 'teardown.sh')
                with open(script_path.localpath, 'w') as w:
                    w.write(teardown_script)
                os.chmod(script_path.localpath, 0o775)
            os.symlink(
                os.path.join(
                    os.path.dirname(__file__),
                    'delocalization.py'
                ),
                os.path.join(self.environment('local')['CANINE_ROOT'], 'delocalization.py')
            )
            return self.finalize_staging_dir(inputs)

    def job_setup_teardown(self, jobId: str, patterns: typing.Dict[str, str]) -> typing.Tuple[str, str]:
        """
        Returns a tuple of (setup script, teardown script) for the given job id.
        Must call after pre-scanning inputs
        """
        job_vars = []
        exports = []
        extra_tasks = [
            'if [[ -d $CANINE_JOB_INPUTS ]]; then cd $CANINE_JOB_INPUTS; fi'
        ]
        compute_env = self.environment('compute')
        for key, val in self.inputs[jobId].items():
            if val.type == 'stream':
                job_vars.append(shlex.quote(key))
                dest = self.reserve_path('jobs', jobId, 'inputs', os.path.basename(os.path.abspath(val.path)))
                extra_tasks += [
                    'if [[ -e {0} ]]; then rm {0}; fi'.format(dest.computepath),
                    'mkfifo {}'.format(dest.computepath),
                    "gsutil {} cat {} > {} &".format(
                        '-u {}'.format(shlex.quote(self.project)) if self.get_requester_pays(val.path) else '',
                        shlex.quote(val.path),
                        dest.computepath
                    )
                ]
                exports.append('export {}="{}"'.format(
                    key,
                    dest.computepath
                ))
            elif val.type == 'download':
                job_vars.append(shlex.quote(key))
                dest = self.reserve_path('jobs', jobId, 'inputs', os.path.basename(os.path.abspath(val.path)))
                extra_tasks += [
                    "if [[ ! -e {2}.fin ]]; then gsutil {0} -o GSUtil:check_hashes=if_fast_else_skip cp {1} {2} && touch {2}.fin; fi".format(
                        '-u {}'.format(shlex.quote(self.project)) if self.get_requester_pays(val.path) else '',
                        shlex.quote(val.path),
                        dest.computepath
                    )
                ]
                exports.append('export {}="{}"'.format(
                    key,
                    dest.computepath
                ))
            elif val.type is None:
                job_vars.append(shlex.quote(key))
                exports.append('export {}={}'.format(
                    key,
                    shlex.quote(val.path.computepath if isinstance(val.path, PathType) else val.path)
                ))
            else:
                print("Unknown localization command:", val.type, "skipping", key, val.path, file=sys.stderr)
        setup_script = '\n'.join(
            line.rstrip()
            for line in [
                '#!/bin/bash',
                'export CANINE_JOB_VARS={}'.format(':'.join(job_vars)),
                'export CANINE_JOB_INPUTS="{}"'.format(os.path.join(compute_env['CANINE_JOBS'], jobId, 'inputs')),
                'export CANINE_JOB_ROOT="{}"'.format(os.path.join(compute_env['CANINE_JOBS'], jobId, 'workspace')),
                'export CANINE_JOB_SETUP="{}"'.format(os.path.join(compute_env['CANINE_JOBS'], jobId, 'setup.sh')),
                'export CANINE_JOB_TEARDOWN="{}"'.format(os.path.join(compute_env['CANINE_JOBS'], jobId, 'teardown.sh')),
                'mkdir -p $CANINE_JOB_INPUTS',
                'mkdir -p $CANINE_JOB_ROOT',
            ] + exports + extra_tasks
        ) + '\ncd $CANINE_JOB_ROOT\n'
        teardown_script = '\n'.join(
            line.rstrip()
            for line in [
                '#!/bin/bash',
                'if [[ -d {0} ]]; then cd {0}; fi'.format(os.path.join(compute_env['CANINE_JOBS'], jobId, 'workspace')),
                # 'mv ../stderr ../stdout .',
                'if which python3 2>/dev/null >/dev/null; then python3 {0} {1} {2} -c {3}; else python {0} {1} {2} -c {3}; fi'.format(
                    os.path.join(compute_env['CANINE_ROOT'], 'delocalization.py'),
                    compute_env['CANINE_OUTPUT'],
                    jobId,
                    ' '.join(
                        '-p {} {}'.format(name, shlex.quote(pattern))
                        for name, pattern in patterns.items()
                    )
                ),
            ]
        )
        return setup_script, teardown_script

    def delocalize(self, patterns: typing.Dict[str, str], output_dir: typing.Optional[str] = None) -> typing.Dict[str, typing.Dict[str, str]]:
        """
        Delocalizes output from all jobs.
        NFSLocalizer does not delocalize files. Data is copied from output dir
        """
        if output_dir is not None:
            warnings.warn("output_dir has no bearing on NFSLocalizer. outputs are available in {}/outputs".format(self.local_dir))
        output_dir = os.path.join(self.local_dir, 'outputs')
        output_files = {}
        for jobId in os.listdir(output_dir):
            start_dir = os.path.join(output_dir, jobId)
            if not os.path.isdir(start_dir):
                continue
            output_files[jobId] = {}
            for outputname in os.listdir(start_dir):
                dirpath = os.path.join(start_dir, outputname)
                if os.path.isdir(dirpath):
                    if outputname not in patterns:
                        warnings.warn("Detected output directory {} which was not declared".format(dirpath))
                    for output in os.listdir(dirpath):
                        fullname = os.path.join(dirpath, output)
                        if fnmatch.fnmatch(fullname, patterns[outputname]) or fnmatch.fnmatch(os.path.relpath(fullname, dirpath), patterns[outputname]):
                            if outputname not in output_files[jobId]:
                                output_files[jobId][outputname] = [os.path.abspath(fullname)]
                            else:
                                output_files[jobId][outputname].append(os.path.abspath(fullname))
        return output_files

    def finalize_staging_dir(self, jobs: typing.Iterable[str], transport: typing.Optional[AbstractTransport] = None) -> str:
        """
        Finalizes the staging directory by building any missing directory trees
        (such as those dropped during transfer for being empty).
        Returns the absolute path of the remote staging directory on the controller node.
        """
        controller_env = self.environment('local')
        if not os.path.isdir(controller_env['CANINE_COMMON']):
            os.mkdir(controller_env['CANINE_COMMON'])
        if not os.path.isdir(controller_env['CANINE_JOBS']):
            os.mkdir(controller_env['CANINE_JOBS'])
        if len(jobs) and not os.path.isdir(controller_env['CANINE_OUTPUT']):
            os.mkdir(controller_env['CANINE_OUTPUT'])
        return self.staging_dir
